.sponsors {
  display: flex;
  justify-content: center;
  flex-wrap: wrap;
  align-items: center;
  margin: 1rem 0;
}

.sponsors > div {
  text-align: center;
  width: 33%;
  padding-bottom: 20px;
}

.sponsors span {
  display: block;
}

@media screen and (max-width: 599px) {
  .sponsors span {
    display: none;
  }
}

.sponsors img {
  width: 65%;
  border-radius: 5px;
}

/*blog post*/
aside.blog {
  display: flex;
  align-items: center;
}

aside.blog img {
  width: 50px;
  height: 50px;
  border-radius: 25px;
  margin-right: 20px;
}

/* Define the company grid layout */

#grid-container {
  width: 100%;
  text-align: center;
}

#company-grid {
  display: inline-block;
  margin: 0 auto;
  gap: 10px;
  align-content: center;
  justify-content: center;
  grid-auto-flow: column;
}

[data-md-color-scheme="slate"] #company-grid {
  background-color: #ffffff;
  border-radius: .5rem;
}

.tile {
  display: flex;
  text-align: center;
  width: 120px;
  height: 120px;
  display: inline-block;
  margin: 10px;
  padding: 5px;
  border-radius: .5rem;
}

.tile img {
  width: 100px;
}
